from collections import namedtuple
from typing import (
    Any,
    List,
)

import matplotlib

# Force matplotlib to not use any Xwindows backend.
matplotlib.use('Agg')

import numpy as np
import matplotlib.pyplot as plt

from astropy import units as u
from astropy.coordinates import (
    Longitude,
    Latitude,
    SkyCoord,
)

from matplotlib.patches import Circle

from gilmenel.instruments import (
    Star,
    BaseInstrument,
    GapInstrument,
)

Circle9 = namedtuple('Circle9', 'x y radius width color label')
Box9 = namedtuple('Box9', 'x y width height rotation color label')


def create_drawing(
    instr: BaseInstrument, stars: List[Star], best_stars: List[Star]
) -> List[Any]:
    # TODO: add hooks to BaseInstrument class so that each instrument and
    # telescope can define its own debug output

    target_ra = instr.target.ra.to_value(u.deg)
    target_dec = instr.target.dec.to_value(u.deg)

    shapes = []

    # setup some fiducials as guides, these are specific to SALT
    # 4.0 arcmin
    shapes.append(Circle9(target_ra, target_dec, 240, 1, 'green', '4 arcmin'))
    # 5.0 arcmin
    shapes.append(Circle9(target_ra, target_dec, 300, 1, 'green', '5 arcmin'))

    # draw inner exclusion distance
    radius = instr.inner_excl_distance.to_value(u.arcmin)
    label = f"{radius} arcmin"
    shapes.append(Circle9(target_ra, target_dec, radius * 60, 1, 'red', label))

    # the actual target/centre of search; in pink so it stands out!
    shapes.append(Circle9(target_ra, target_dec, 10, 3, 'magenta', "target"))

    # a line near the target to make sure the slit gap width is OK
    if isinstance(instr, GapInstrument):
        gap_width = instr.slit_gap_radius.to_value(u.arcmin) * 2
        pa = instr.slit_gap_angle.to_value(u.rad)
        shapes.append(Box9(target_ra, target_dec, gap_width, 10, pa, 'red', ""))

    # TODO: make these colours editable too
    merit_colours = {
        -1: "grey",  # not processed
        0: "red",  # out of range
        1: "orange",  # nearby stars
        2: "yellow",  # too dim
        3: "yellow",  # too bright
        4: "green",  # good candidate
    }

    for s in stars:
        if s.merit > -1:  # exclude faint stars that were not evaluated
            # the colour reflects the star's suitability
            colour = merit_colours[s.merit]
            shapes.append(
                Circle9(
                    s.ra.to_value(u.deg),
                    s.dec.to_value(u.deg),
                    2,
                    1,
                    colour,
                    f"{s.g_mag:.2f}",
                )
            )

    # indicate the best stars
    for s in best_stars:
        if s is not None:
            shapes.append(
                Circle9(
                    s.ra.to_value(u.deg),
                    s.dec.to_value(u.deg),
                    10,
                    2,
                    'cyan',
                    "best star",
                )
            )

    return shapes


def produce_ds9_region(
    instr: BaseInstrument, stars: List[Star], best_stars: List[Star]
) -> str:
    # create a ds9 region to help visualise things - for testing purposes...
    # note this must be written to disk as a region file that must be loaded
    # from file by ds9 later with a system command

    def draw_circle(r, label, c, x, y, width):
        rline = (
            f"circle {x:.6f}d {y:.6f}d {r}\" "
            f"# text={{{label}}} color={c} width={width}\n"
        )
        return rline

    shapes = create_drawing(instr, stars, best_stars)

    output = "global color=white\n"
    output += "fk5\n"

    # draw shapes
    for sh in shapes:
        if isinstance(sh, Circle9):
            output += draw_circle(sh.radius, sh.label, sh.color, sh.x, sh.y, sh.width)

        elif isinstance(sh, Box9):
            output += (
                f"box {sh.x} {sh.y} {sh.width}' {sh.height}' {sh.rotation}r #"
                f" fill=0 color={sh.color}\n"
            )

    return output


def produce_png(instr: BaseInstrument, stars: List[Star], best_stars: List[Star]):
    def red_or_blue(star):
        # calculate an approximate colour for the star
        # ideal: C = BP - RP -> 0.5 = blue, 2.0 = green/grey, 3.5 = red

        if star.rp_mag:
            return star.g_mag - star.rp_mag - 0.9  # more positive = more red

        return 0

    def equitorial(target, coords):
        ''' Convert ICRS coordinates in degrees to
            equitorially equivalent coordinates
        '''
        ra = (coords.ra - target.ra) * np.cos(coords.dec.radian)
        dec = coords.dec - target.dec

        return ra.to_value(u.deg), dec.to_value(u.deg)

        #     return ra.ra.to_value(u.deg) / np.cos(
        #         np.radians(ra.dec.to_value(u.deg)))

        # return ra / np.cos(np.radians(target_dec))

    def icrs(target, coords):
        ''' Convert equitorially equivalent coordinates in degrees to
            ICRS coordinates
        '''
        ra = coords.ra / np.cos(coords.dec.radian) + target.ra
        dec = coords.dec + target.dec

        return ra.wrap_at(360 * u.deg).degree, dec.degree

    target = instr.target

    # get brightest star
    brightest = min(s.g_mag for s in stars)

    # discard stars that are dimmer by x magnitudes
    stars = [s for s in stars if s.g_mag <= brightest + 8]

    ra, dec = zip(*[equitorial(target, s) for s in stars])
    colour = [red_or_blue(s) for s in stars]
    g_mag = [s.g_mag for s in stars]
    merit = [s.merit for s in stars]

    # 'brightness' for star size on scatter plot
    # 20 -> 1
    # 15 -> 5
    # 10 -> 25
    # = 5 ** ((20 - g) / 5)
    brightness = [(3 ** ((20 - s.g_mag) / 5)) ** 2.0 for s in stars]
    halo = [(3 ** ((20 - s.g_mag) / 5)) ** 2.5 for s in stars]

    fig, ax = plt.subplots(figsize=(12.5, 10))
    ax.set_facecolor('#111111')

    # halo
    ax.scatter(
        ra,
        dec,
        c=colour,
        s=halo,
        alpha=0.4,
        marker='o',
        edgecolors="none",
        cmap=plt.cm.RdYlBu_r,
        vmin=-1.5,
        vmax=1.5,
    )
    # centers
    sc = ax.scatter(
        ra,
        dec,
        c=colour,
        s=brightness,
        alpha=0.9,
        marker='D',
        # marker="$\u25b2$",
        edgecolors="none",
        cmap=plt.cm.RdYlBu_r,
        vmin=-1.5,
        vmax=1.5,
        label="Guide Stars",
    )

    # TODO: make these colours editable too
    merit_colours = {
        -1: "grey",  # not processed
        0: "#FF0000",  # out of range
        1: "#FF5500",  # nearby stars
        2: "#C0CA33",  # too dim
        3: "#FFFF00",  # too bright
        4: "#00FF00",  # good candidate
    }

    text_offset = 4 * 10 ** -3
    # the colour reflects the star's suitability
    merit_colour = [merit_colours[s.merit] for s in stars]
    for r, d, g, m in zip(ra, dec, g_mag, merit):
        if m > -1:
            ax.annotate(
                f"{g:.2f}", (r, d + text_offset), color=merit_colours[m], ha='center',
            )

    # add fiducials

    # transform to equatorial equivalent coordinates
    one_arcmin = 1 / 60

    # SCAM
    ax.add_artist(Circle((0, 0), 5 * one_arcmin, color='#00ff00', fill=False,))
    # RSS
    ax.add_artist(Circle((0, 0), 4 * one_arcmin, color='#00ff00', fill=False,))
    # Target Exclusion
    exclusion = instr.inner_excl_distance.to_value(u.arcmin)
    ax.add_artist(Circle((0, 0), exclusion * one_arcmin, color='#990000', fill=False,))
    # Slit exclusion
    if isinstance(instr, GapInstrument):
        # TODO: add slit rectangle that rotates with PA
        pass
    # Target
    ax.add_artist(Circle((0, 0), 10 / 60 * one_arcmin, color='#ff00ff', fill=False,))

    # Guide Stars
    text_offset = 4 * 10 ** -3
    for r, d in [equitorial(instr.target, s) for s in best_stars]:
        ax.add_artist(Circle((r, d), 10 / 60 * one_arcmin, color='cyan', fill=False,))
        ax.annotate(
            "Guide Star", (r + text_offset, d), color='cyan', ha='right',
        )

    # set axis limits
    ax.set_xlim(5.25 * one_arcmin, -5.25 * one_arcmin)
    ax.set_ylim(-5.25 * one_arcmin, 5.25 * one_arcmin)
    # set ticks
    ra_ticks = [
        Longitude(icrs(target, SkyCoord(t * u.deg, 0 * u.deg))[0], u.deg).to_string(
            unit=u.hour, precision=0
        )
        for t in ax.get_xticks()
    ]
    ax.set_xticklabels(ra_ticks)
    dec_ticks = [
        Latitude(icrs(target, SkyCoord(0 * u.deg, t * u.deg))[1], u.deg).to_string(
            unit=u.deg, precision=0
        )
        for t in ax.get_yticks()
    ]
    ax.set_yticklabels(dec_ticks)
    # label axes
    ax.set_xlabel(r'$RA$', fontsize=12)
    ax.set_ylabel(r'$Dec$', fontsize=12)
    ax.set_title("Guide Stars (SCAM Display)")
    # add colourbar
    cbar = plt.colorbar(sc)
    cbar.ax.set_ylabel("Source Colour")
    # output figure
    fig.tight_layout()
    plt.savefig("./charts/current.png", bbox_inches='tight')
    plt.close()



#!.venv/bin/python3

import config

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from models import (
    Base,
    GAIA_DR2,
)

engine = create_engine(f"sqlite:///{config.catalogue_db_path}", echo=True)

Base.metadata.create_all(engine)

Session = sessionmaker(bind=engine)
session = Session()

session.commit()

session.add(GAIA_DR2)
session.commit()

session.execute('PRAGMA journal_mode=WAL;')



#!.venv/bin/python3

import numpy as np

from collections import deque

from astropy import units as u
from astropy.coordinates import SkyCoord
from astroquery.vizier import Vizier

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import NullPool

from tqdm import tqdm, trange

from models import (
    GAIA_DR2,
    Source,
)

session = None
waiting_sources = deque()


def open_session():
    global session

    engine = create_engine('sqlite:///catalogue/gaia_dr2_salt.db', poolclass=NullPool,)

    Session = sessionmaker(bind=engine)
    session = Session()

    # use local Vizier mirror
    Vizier.VIZIER_SERVER = 'vizieridia.saao.ac.za'


def download_sources(target: SkyCoord, faint_limit: float):
    # Query the GAIA DR2 vizier catalogue
    # (vizier catalogue identifier: I/345/gaia2)
    # The '+' in "+Gmag" sorts the list by the brightest first

    # The proper motion (pmRA and pmDE) filter is needed to remove some stars
    # that may be fast moving e.g. block_id 78169 is an interesting test!
    # -there's a 13.5 mag star with high proper motion!
    # This is not so much a problem *now*, soon after GAIA catalogue was made,
    # but is more important for future proofing this code...
    # Limits of +-50mas/yr are selected in RA and Dec.
    vquery = Vizier(
        columns=['DR2Name', 'RA_ICRS', 'DE_ICRS', 'Gmag', 'RPmag',],
        column_filters={
            "Gmag": (f"<={faint_limit:.2f}"),
            "pmRA": ("> -50 && < 50"),
            "pmDE": ("> -50 && < 50"),
        },
        row_limit=-1,
    )
    vquery.TIMEOUT = 120  # 2 mins
    target.field = 'icrs'
    results = vquery.query_region(
        target,
        width=1.0 * np.cos(target.dec) * u.deg,
        height=1.0 * u.deg,
        catalog="I/345/gaia2",
        cache=False,
    )[0]

    # if we find no stars in the field, raise an error
    if len(results) == 0:
        raise Exception("No stars found in field")
    else:
        # for r in results:
        #     print(type(r['Source']), r['Source'])
        return results


def write_to_db(sources):
    global session

    for s in tqdm(sources, mininterval=1.0, leave=False):
        source = Source(
            source=s['DR2Name'][9:],
            catalog=GAIA_DR2,
            ra=float(s['RA_ICRS']),
            dec=float(s['DE_ICRS']),
            g_mag=float(s['Gmag']),
            rp_mag=(
                float(s['RPmag'])
                if not isinstance(s['RPmag'], np.ma.core.MaskedConstant)
                else np.nan
            ),
        )
        session.add(source)

    session.commit()


def queue_to_write(sources):
    global session
    global waiting_sources

    for s in sources:
        waiting_sources.append(
            Source(
                source=s['DR2Name'][9:],
                catalog=GAIA_DR2,
                ra=float(s['RA_ICRS']),
                dec=float(s['DE_ICRS']),
                g_mag=float(s['Gmag']),
                rp_mag=(
                    float(s['RPmag'])
                    if not isinstance(s['RPmag'], np.ma.core.MaskedConstant)
                    else np.nan
                ),
            )
        )
    if len(waiting_sources) >= 30000:
        for _ in trange(len(waiting_sources), leave=False):
            session.add(waiting_sources.popleft())
        session.commit()


def write_checkpoint():
    # add remaining sources
    for _ in trange(len(waiting_sources), leave=False):
        session.add(waiting_sources.popleft())
    session.commit()

    # commit all WAL changes
    session.execute('PRAGMA wal_checkpoint;')


def download(target: SkyCoord, faint_limit: float = 21):
    sources = download_sources(target, faint_limit)
    print(f"[{target.ra}, {target.dec}] returned {len(sources)} sources")
    write_to_db(sources)


if __name__ == '__main__':
    open_session()
    target = SkyCoord(193.5 * u.deg, -60.5 * u.deg)  # Jewel Box
    # target = SkyCoord(165 * u.deg, -56 * u.deg)  # debug
    # target = SkyCoord(165.5 * u.deg, -55.5 * u.deg)  # debug
    # target = SkyCoord(0 * u.deg, 0 * u.deg)  # Null Island
    # target = SkyCoord(120 * u.deg, -30 * u.deg)  # somewhere South
    download(target, faint_limit=18.5)



import math

# models created from: http://bytefish.de/blog/first_steps_with_sqlalchemy/

from sqlalchemy import (
    Column,
    Float,
    ForeignKey,
    Index,
    inspect,
    Integer,
    String,
)
from sqlalchemy.ext.declarative import (
    as_declarative,
    declarative_base,
)
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship


# returns a row as a dictionary
@as_declarative()
class Base:
    def _asdict(self):
        return {c.key: getattr(self, c.key) for c in inspect(self).mapper.column_attrs}


class Catalog(Base):
    __tablename__ = 'catalogs'

    catalog_id = Column(Integer(), unique=True, primary_key=True)
    name = Column(String(20))
    prefix = Column(String(20))


class Source(Base):

    __tablename__ = 'sources'

    # Columns
    # -------
    # source
    # ra
    # dec
    # g_mag
    # rp_mag

    source_id = Column(Integer(), unique=True, primary_key=True)
    catalog_id = Column(Integer(), ForeignKey('catalogs.catalog_id'))
    catalog = relationship("Catalog")
    source = Column(String(20))
    ra = Column(Float())
    dec = Column(Float())
    g_mag = Column(Float())
    rp_mag = Column(Float())

    @hybrid_property
    def ra_square(self):
        return int(math.floor(self.ra))

    @hybrid_property
    def dec_square(self):
        return int(math.floor(self.dec)) + 90

    @hybrid_property
    def ra_deg_dist(self):
        return int(math.floor(self.ra % 1 * 100))

    @hybrid_property
    def dec_deg_dist(self):
        return int(math.floor(self.dec % 1 * 100))

    @hybrid_property
    def source_name(self):
        return f"{self.catalog.prefix}{self.source}"

    def __repr__(self):
        return (
            f"<Source ("
            f"id='{self.source_id}', "
            f"source='{self.source}', "
            f"ra='{self.ra}', "
            f"dec='{self.dec}', "
            f"g_mag='{self.g_mag}', "
            f"rp_mag='{self.rp_mag}'"
            f")>"
        )

    def create_indexes(engine):
        # indexes
        sources_ra_index = Index('sources_ra_idx', Source.ra)
        sources_dec_index = Index('sources_dec_idx', Source.dec)
        sources_gmag_index = Index('sources_gmag_idx', Source.g_mag)

        sources_ra_index.create(bind=engine)
        sources_dec_index.create(bind=engine)
        sources_gmag_index.create(bind=engine)


# constants
GAIA_DR2 = Catalog(name="Gaia DR2", prefix="Gaia DR2 ")



from typing import List

from astropy import units as u
from astropy.coordinates import SkyCoord
from astroquery.vizier import Vizier

from gilmenel import astroquery as asterquery
from gilmenel.exceptions import NoStarsFoundError
from gilmenel.diagrams import (
    produce_ds9_region,
    produce_png,
)
from gilmenel.instruments import (
    Star,
    BaseInstrument,
    GapInstrument,
)

Catalog = None


def init(use_local_catalogue=False, catalog_url=None):
    global Catalog

    if use_local_catalogue:
        Catalog = asterquery.Local
        asterquery.catalog_url = catalog_url
    else:
        Catalog = Vizier
        print("Using remote Vizier catalogue")


def view_sky(instr: BaseInstrument, max_stars: int = -1) -> List[Star]:
    # Query the GAIA DR2 vizier catalogue (catalogue identifier: I/345/gaia2)
    # The '+' in "+Gmag" sorts the list by the brightest first, which is useful
    # for us later when it comes to picking the best candidate stars.
    # The '_r' column gives a handy radial distance of the star from the search
    # position ra_deg, dec_deg

    # The proper motion (pmRA and pmDE) filter is needed to remove some stars
    # that may be fast moving, e.g. block_id 78169 is an interesting test!
    # - There's a 13.5 mag star with high proper motion!
    # This is not so much a problem *now*, soon after GAIA catalogue was made,
    # but is more important for future proofing this code...
    # Limits of +-50mas/yr are selected in RA and Dec.
    global Catalog

    query = Catalog(
        columns=["_r", 'Source', 'RA_ICRS', 'DE_ICRS', '+Gmag', 'RPmag'],
        column_filters={
            "Gmag": (f"<={instr.faint_limit + 2:.2f}"),
            "pmRA": ("> -50 && < 50"),
            "pmDE": ("> -50 && < 50"),
        },
        # column_filters={"Gmag":(">=13.0 && <=17.0")},
        row_limit=max_stars,
    )
    results = query.query_region(
        instr.target, radius=instr.instr_fov, catalog="I/345/gaia2"
    )[0]

    # if we find no stars in the field, raise an error
    if len(results) == 0:
        raise NoStarsFoundError(
            f"No stars returned in field "
            f"'{instr.target.to_string(style='hmsdms')}' "
            f"with radius {instr.instr_fov:.2f}"
        )
    else:
        stars = []
        for row in results:
            stars.append(
                Star(
                    ra=row['RA_ICRS'] * u.deg,
                    dec=row['DE_ICRS'] * u.deg,
                    radius=row['_r'] * u.arcmin,
                    g_mag=row['Gmag'],
                    rp_mag=row['RPmag'],
                )
            )

    return stars


def find_best_stars(instr: BaseInstrument, stars: List[Star]) -> List[Star]:
    '''
    Return the best stars as defined by the instrument's best_stars function
    '''
    stars = instr.filter(stars)

    return instr.best_stars(stars)



#!.venv/bin/python3

import os.path as path
import pickle
import threading
import sys

import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.ticker as ticker

import numpy as np

import config

from concurrent.futures import ProcessPoolExecutor as PoolExecutor
from itertools import tee
from time import strftime
from tqdm import tqdm
from typing import List, Tuple

from sqlalchemy import (
    and_,
    create_engine,
    func,
)
from sqlalchemy.orm import sessionmaker

from models import Source

session = None


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def open_session():
    global session

    engine = create_engine(f"sqlite:///{config.catalogue_db_path}")

    Session = sessionmaker(bind=engine)  # noqa
    session = Session()


def query_source_ids(spread: Tuple[int, int]):
    start, stop = spread

    star_grid_sky = np.zeros((180, 360))  # whole sky
    star_grid_deg = np.zeros((100, 100))  # per degree

    query = session.query(Source)
    query = query.filter(Source.source_id >= start)
    query = query.filter(Source.source_id < stop)

    for i, s in enumerate(query.yield_per(128)):
        # increment count in grid of source locations on sky
        star_grid_sky[s.dec_square][s.ra_square] += 1

        # increment count in grid of source locations per deg
        star_grid_deg[s.dec_deg_dist][s.ra_deg_dist] += 1

    return star_grid_sky, star_grid_deg


def query_grid():
    # ID_MAX = 104259626
    ID_MAX = (
        session.query(Source.source_id)
        .order_by(Source.source_id.desc())
        .limit(1)
        .one()[0]
    )  # index of last record in table
    ID_STEP = 10 ** 4
    WORKERS = 12

    star_ids = ((p1, p2) for p1, p2 in pairwise(range(0, ID_MAX, ID_STEP)))
    star_grid_sky = np.zeros((180, 360))
    star_grid_deg = np.zeros((100, 100))

    print(f"Expecting {ID_MAX:,} sources")

    try:
        # create a thread pool of WORKERS threads
        with PoolExecutor(max_workers=WORKERS) as executor:
            # distribute all star_ids between the workers in the pool
            # grid is the response of query_source_ids
            for grid_sky, grid_deg in tqdm(
                executor.map(query_source_ids, star_ids),
                total=round(ID_MAX / ID_STEP),
                desc="Loading sources",
                unit=" kilostars",
            ):
                star_grid_sky = [
                    [s + g for s, g in zip(s_row, g_row)]
                    for s_row, g_row in zip(star_grid_sky, grid_sky)
                ]

                star_grid_deg = [
                    [s + g for s, g in zip(s_row, g_row)]
                    for s_row, g_row in zip(star_grid_deg, grid_deg)
                ]

    except KeyboardInterrupt:
        print("Keyboard interrupt - query incomplete")

        return star_grid_sky, star_grid_deg

    return star_grid_sky, star_grid_deg


def plot_stars(stars):
    matfig = plt.figure(figsize=(20, 10))
    im = plt.matshow(
        stars,
        aspect='equal',
        origin='lower',
        extent=[0, 360, -90, 90],
        norm=colors.LogNorm(vmin=10 ** 2, vmax=10 ** 6),
        fignum=matfig.number,
    )
    plt.colorbar(im).ax.set_ylabel('Source Density per deg²', rotation=270)
    plt.gca().xaxis.tick_bottom()
    plt.gca().yaxis.set_major_locator(ticker.MultipleLocator(30))
    # x-axis relabelling from degrees to hours
    plt.gca().set_xlim(360, 0)
    plt.gca().set_xticks(np.arange(0, 360 + 1, 360 / 12))
    plt.gca().set_xticklabels(np.arange(0, 24 + 1, 2))
    plt.xlabel('RA [hours]')
    # add second x-axis in degrees
    ax2 = plt.gca().secondary_xaxis('top', functions=(lambda x: x, lambda x: x),)
    ax2.set_ticks(np.arange(0, 360 + 1, 30))
    ax2.set_xlabel('RA [degrees]')
    # continue with y-axis
    plt.ylabel('Dec [degrees]')
    plt.title('Downloaded Catalogue Source Density (Gmag < 18.5)', pad=20)
    plot_path = (
        path.splitext(config.catalogue_db_path)[0]
        + f'_sky_{strftime("%Y-%m-%dT%H-%M-%S")}.png'
    )
    plt.savefig(plot_path, bbox_inches='tight')
    plt.show()
    plt.close()


def plot_dist(stars):
    matfig = plt.figure(figsize=(10, 10))
    im = plt.matshow(
        stars,
        aspect='equal',
        origin='lower',
        # extent=[0, 360, -90, 90],
        norm=colors.LogNorm(vmin=10 ** 0, vmax=10 ** 6),
        fignum=matfig.number,
    )
    plt.colorbar(im).ax.set_ylabel('Source Density per deg²', rotation=270)
    plt.gca().xaxis.tick_bottom()
    plt.gca().yaxis.set_major_locator(ticker.MultipleLocator(10))
    # x-axis relabelling from degrees to hours
    # plt.gca().set_xlim(360, 0)
    # plt.gca().set_xticks(np.arange(0, 360 + 1, 360 / 12))
    # plt.gca().set_xticklabels(np.arange(0, 24 + 1, 2))
    plt.xlabel('RA [degrees x10^2]')
    # add second x-axis in degrees
    ax2 = plt.gca().secondary_xaxis('top', functions=(lambda x: x, lambda x: x),)
    # ax2.set_ticks(np.arange(0, 360 + 1, 30))
    ax2.set_xlabel('RA [degrees x10^2]')
    # continue with y-axis
    plt.ylabel('Dec [degrees x10^2]')
    plt.title('Downloaded Catalogue Source Density (Gmag < 18.5)', pad=20)
    plot_path = (
        path.splitext(config.catalogue_db_path)[0]
        + f'_deg_{strftime("%Y-%m-%dT%H-%M-%S")}.png'
    )
    plt.savefig(plot_path, bbox_inches='tight')
    plt.show()
    plt.close()


if __name__ == '__main__':

    debug = False
    if len(sys.argv) > 1 and sys.argv[1] == '-d':
        debug = True

    pickle_path = path.splitext(config.catalogue_db_path)[0] + "_summary.pickle"

    if debug:
        print("Loading pickle")
        stars_sky, stars_deg = pickle.load(open(pickle_path, 'rb'))
    else:
        print("Querying database...")
        open_session()
        stars_sky, stars_deg = query_grid()
        # pickle to save results from expensive query
        pickle.dump((stars_sky, stars_deg), open(pickle_path, 'wb'))

    print("Plotting stars...")
    plot_stars(stars_sky)
    plot_dist(stars_deg)
    print("Complete.")



class AsteriaError(Exception):
    """Base class for exceptions in this module."""

    pass


class CatalogUnavailableError(AsteriaError):
    """Raised when the requested catalogue is not available."""

    pass


class NoStarsFoundError(AsteriaError):
    """Raised when no stars are returned in field."""

    pass



#!.venv/bin/python3

from sqlalchemy import create_engine

import config
from models import Base

engine = create_engine(f"sqlite:///{config.catalogue_db_path}", echo=True)

Base.metadata.drop_all(engine)



import abc
import re

from typing import (
    List,
    Optional,
)

import numpy as np

from astropy import units as u
from astropy.coordinates import SkyCoord


class Star(object):
    def __init__(
        self,
        ra: u.deg,
        dec: u.deg,
        radius: u.arcmin,
        g_mag: float,
        rp_mag: Optional[float] = None,
        merit: Optional[int] = None,
    ):
        self.sky_coord = SkyCoord(ra, dec, frame='icrs')
        self.ra = self.sky_coord.ra
        self.dec = self.sky_coord.dec

        self.radius = radius  # arcmin
        self.g_mag = g_mag
        self.rp_mag = rp_mag

        # the stars coordinates in the instrument's reference frame
        self.instr_coord = None

        # at first, all stars have a merit of -1
        # as they pass each filter, the merit value increases
        # this will allow for operations in marginal conditions
        if merit is None:
            self.merit = -1
        else:
            self.merit = merit

    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(other, Star):
            return (
                u.isclose(self.ra, other.ra, rtol=1e-6)
                and u.isclose(self.dec, other.dec, rtol=1e-6)
                and u.isclose(self.radius, other.radius, rtol=1e-2)
                and self.g_mag == other.g_mag
                and self.rp_mag == other.rp_mag
                and self.merit == other.merit
            )
        return False  # pragma: no cover

    def __str__(self):
        return (
            f"ra={self.ra:.6f}, "
            f"dec={self.dec:.6f}, "
            f"radius={self.radius:.3f}, "
            f"g_mag={self.g_mag:.2f}, "
            f"{f'rp_mag={self.rp_mag:.2f}, ' if self.rp_mag else ''}"
            f"merit={self.merit}"
        )  # pragma: no cover

    def __repr__(self):
        return re.sub(
            r'0+ ',
            '0 ',
            f"\nStar({self.ra.to_value(u.deg):.6f} * d,"
            f" {self.dec.to_value(u.deg):.6f} * d,"
            f" {self.radius.to_value(u.arcmin):.3f} * m,"
            f" {self.g_mag},"
            f"{f' rp_mag={self.rp_mag:.2f},' if self.rp_mag else ''}"
            f" merit={self.merit})",
        )  # pragma: no cover


class BaseInstrument(metaclass=abc.ABCMeta):
    def __init__(
        self,
        name: str,
        instr_fov: u.arcmin,
        inner_excl_distance: u.arcmin,
        nearby_limit: u.arcsec,
        bright_limit: float,
        faint_limit: float,
    ):
        self.name = name
        self.instr_fov = instr_fov  # arcminutes radius
        self.inner_excl_distance = inner_excl_distance  # arcminutes radius
        self.nearby_limit = nearby_limit  # arcseconds diameter
        self.bright_limit = bright_limit
        self.faint_limit = faint_limit
        self.inner_excl_shape = 'circle'

        self._target = None  # set by instr.point_to(SkyCoord(ra=x, dec=y))

    def __getattribute__(self, name):
        if name == 'target':
            return self._target
        else:
            return object.__getattribute__(self, name)

    # args and kwargs allow easier overloading of this function
    def point_to(self, target, *args, **kwargs):
        self._target = target
        # set instrument frame from target position
        self.instr_frame = target.skyoffset_frame()

    def star_available(self, star: Star) -> bool:
        '''Check if star location falls within allowable geometry'''

        # set star coordinates relative to instrument frame
        star.instr_coord = star.sky_coord.copy()
        star.instr_coord = star.instr_coord.transform_to(self.instr_frame)

        # Check if star falls within FOV
        if star.radius > self.instr_fov:
            return False

        # Check if star falls within exclusion zone
        if self.inner_excl_shape == 'circle':
            if star.radius <= self.inner_excl_distance:
                return False
        elif self.inner_excl_shape == 'square':
            if (
                star.instr_coord.lon <= self.inner_excl_distance
                and star.instr_coord.lon >= -self.inner_excl_distance
                and star.instr_coord.lat <= self.inner_excl_distance
                and star.instr_coord.lat >= -self.inner_excl_distance
            ):
                return False
        else:
            raise NotImplementedError

        return True

    def filter_geometry(self, stars: List[Star]) -> List[Star]:
        for s in stars:
            if self.star_available(s):
                s.merit = 1
            else:
                s.merit = 0  # mark that star has been checked

        return stars

    def filter_nearby_pairs(self, stars: List[Star]) -> List[Star]:
        for s in stars:

            def s_is_close_to(t):
                return (
                    t.g_mag < s.g_mag + 2
                    and abs(t.ra - s.ra)  # t mag brighter than s mag -2
                    <= self.nearby_limit
                    and abs(t.dec - s.dec) <= self.nearby_limit
                )

            nearby_stars = any(  # any returns True at first test star found
                True for t in stars if s_is_close_to(t) and s is not t
            )  # t -> test star

            if nearby_stars is False:
                s.merit = 2

        return stars

    def filter_magnitudes(self, stars: List[Star]) -> List[Star]:
        for s in stars:
            # remove faint stars
            if s.g_mag > self.faint_limit:
                pass  # leave faint stars at their current merit value

            # remove bright stars
            elif s.g_mag < self.bright_limit:
                s.merit = 3

            # just right
            else:
                s.merit = 4

        return stars

    def criteria(self, stars: List[Star]) -> bool:
        '''
        For filtering stars, return True once enough stars have been found
        '''
        self.filter_geometry(stars)
        self.filter_nearby_pairs([s for s in stars if s.merit == 1])
        self.filter_magnitudes([s for s in stars if s.merit == 2])

        # check if enough stars have been found
        if len([s for s in stars if s.merit >= 4]) > 10:
            return True

        return False

    def filter(self, stars: List[Star]) -> List[Star]:
        '''
        Filter stars in layers of magnitude, each call to criteria will be
        given more stars.

        Two magnitudes of additional stars are included to make sure that
        dim, yet visible, stars that are too close are rejected later.
        '''
        brightest = min(s.g_mag for s in stars)
        faintest = max(s.g_mag for s in stars)

        # make sure to include at least one magnitude if brightest == faintest
        test_magnitudes = np.arange(brightest, max(brightest + 1, faintest - 2))

        star_subset = []
        for mag in test_magnitudes:
            star_subset = [s for s in stars if s.g_mag <= mag + 2]

            if self.criteria(star_subset) is True:
                # if enough stars have been found
                break

        return star_subset

    @abc.abstractmethod
    def best_stars(self, stars: List[Star]) -> List[Star]:
        '''Return the best guide stars from the given selection'''


class GapInstrument(BaseInstrument):
    def __init__(
        self,
        name: str,
        instr_fov: u.arcmin,
        inner_excl_distance: u.arcmin,
        nearby_limit: u.arcsec,
        bright_limit: float,
        faint_limit: float,
        slit_gap_radius: u.arcmin,
        slit_gap_angle: u.deg,
    ):
        super().__init__(
            name,
            instr_fov,
            inner_excl_distance,
            nearby_limit,
            bright_limit,
            faint_limit,
        )
        self.slit_gap_radius = slit_gap_radius  # arcmin
        self.slit_gap_angle = slit_gap_angle  # degrees

    def point_to(self, target, pa=None):
        # set target if available
        if target is not None:
            super().point_to(target)

        # set position angle if available
        if pa is not None:
            self.slit_gap_angle = pa

        # set instrument frame from target position
        self.instr_frame = self.target.skyoffset_frame(rotation=self.slit_gap_angle)

    def star_available(self, star: Star) -> bool:
        '''Check if star location falls within allowable geometry'''

        # first call parent method
        result = super().star_available(star)
        if result is False:
            return result

        # Check if star falls within a vertical gap of distance slit_gap_radius
        ra = (star.instr_coord.lon.to_value(u.deg) + 180) % 360 - 180

        if abs(ra) <= self.slit_gap_radius.to_value(u.deg):
            return False

        return True

    def best_stars(self, stars: List[Star]) -> List[Star]:
        '''Return the best guide stars from the given selection'''
        raise NotImplementedError



from astropy import units as u

from gilmenel import gilmenel


class GuiderSingle(gilmenel.BaseInstrument):
    def best_stars(self, stars):
        return [s for s in stars if s.merit >= 4][:1]


fif = GuiderSingle(
    'FIF',  # instr_name
    5 * u.arcmin - 20 * u.arcsec,  # instr_fov, arcminutes radius
    1.5 * u.arcmin,  # inner_excl_distance, arcminutes
    10 * u.arcsec,  # probe_fov, arcseconds
    8,  # bright limit
    14,  # faint limit
)


class GuiderDual(gilmenel.GapInstrument):
    def _score_star(self, star):
        '''
        rank each star linearly by magnitude and distance from exclusion zone
        favours bright stars at the edge of the field
        '''
        radius = (star.radius - self.inner_excl_distance) / (
            self.instr_fov - self.inner_excl_distance - self.nearby_limit
        )
        mag = 1 - (star.g_mag - self.bright_limit) / (
            self.faint_limit - self.bright_limit
        )
        return radius * mag

    def _group_by_side(self, stars):
        '''
        groups stars into left and right of gap
        '''
        stars_left = [
            s for s in stars if s.merit >= 4 and s.instr_coord.lon.to_value(u.deg) < 0
        ]
        stars_right = [
            s for s in stars if s.merit >= 4 and s.instr_coord.lon.to_value(u.deg) > 0
        ]

        return stars_left, stars_right

    def _furthest_pair(self, stars_left, stars_right):
        '''
        compare distances between the top 3 stars on each list
        to get longest distance between a star pair
        '''
        furthest_pair = (0, None, None)  # seperation, s, t
        for s in stars_left[:3]:
            for t in stars_right[:3]:
                sep = s.instr_coord.separation(t.instr_coord)
                if sep > furthest_pair[0]:
                    furthest_pair = sep, s, t

        return furthest_pair

    def criteria(self, stars):
        self.filter_geometry(stars)
        self.filter_nearby_pairs([s for s in stars if s.merit == 1])
        self.filter_magnitudes([s for s in stars if s.merit == 2])

        # check if enough stars have been found
        if len([s for s in stars if s.merit >= 4]) > 6:
            return True

    def best_stars(self, stars):
        # split all stars into left and right
        stars_left, stars_right = self._group_by_side(stars)

        # sort in place by star score
        stars_left.sort(key=lambda s: self._score_star(s), reverse=True)
        stars_right.sort(key=lambda s: self._score_star(s), reverse=True)

        # choose the stars with the greatest seperation
        _, star_left, star_right = self._furthest_pair(stars_left, stars_right)

        return star_left, star_right


pfgs = GuiderDual(
    'PFGS',  # instr_name
    5 * u.arcmin - 22 * u.arcsec,  # instr_fov, arcminutes radius
    1.0 * u.arcmin,  # inner_excl_distance, arcminutes
    11 * u.arcsec,  # probe_fov, arcseconds
    9,  # bright limit
    16.5,  # faint limit
    0.65 * u.arcmin,  # arcmin
    0 * u.deg,  # degrees
)
pfgs.inner_excl_shape = 'square'



#!.venv/bin/python3

import itertools
import sys

import click
import requests

from astropy import units as u
from astropy.coordinates import SkyCoord

from tqdm import tqdm

import download_stars

SUCCESS_FILE = "catalogue/succeeded.tsv"
FAILED_FILE = "catalogue/failed.tsv"


def download_degree2(target):
    attempts = 0
    success = False
    sources = None

    while success is False:
        try:
            sources = download_stars.download_sources(target, faint_limit=18.5)
            success = True

            with open(SUCCESS_FILE, 'a') as f:
                # append coords to tsv file
                f.write(f"{target.ra}\t{target.dec}\t{len(sources)}\n")
        except requests.exceptions.ConnectionError:
            # try again
            attempts += 1
            tqdm.write(f"ConnectionError, occurrance {attempts} - retrying")
        except requests.exceptions.ReadTimeout:
            # try again
            attempts += 1
            tqdm.write(f"ReadTimeout, occurrance {attempts} - retrying")
        except Exception as e:

            with open(FAILED_FILE, 'a') as f:
                # append coords to tsv file
                f.write(f"{target.ra}\t{target.dec}\t{e}\n")

    return sources


def main(start_ra=0, start_dec=-90, size=360 * 180):
    RA_POS = 360
    RA_NEG = 0

    DEC_POS = 11
    DEC_NEG = -76

    start_ra = max(start_ra, RA_NEG)
    start_dec = max(start_dec, DEC_NEG)

    # create grid covering telescope range
    full_grid = (
        (ra, dec) for ra in range(RA_NEG, RA_POS) for dec in range(DEC_NEG, DEC_POS)
    )

    # remove section excluded by starting ra and dec
    partial_grid = (
        (ra, dec)
        for ra, dec in full_grid
        if (ra > start_ra or (ra == start_ra and dec >= start_dec))
    )

    # create SkyCoords from remaining grid
    targets = (
        SkyCoord((ra + 0.5) * u.deg, (dec + 0.5) * u.deg) for ra, dec in partial_grid
    )

    size = min(
        ((RA_POS - start_ra - 1) * (DEC_POS - DEC_NEG) + (DEC_POS - start_dec)), size
    )

    print(f"Total patches: {size}")

    with open(SUCCESS_FILE, 'w') as f:
        # clear success file
        f.write(f"RA\tDec\tSources\n")

    with open(FAILED_FILE, 'w') as f:
        # clear failure file
        f.write(f"RA\tDec\tReason\n")

    download_stars.open_session()

    # limit size of iterator based on size
    targets = itertools.islice(targets, size)

    print("Executing...")

    for t in tqdm(targets, total=size, desc="Downloads", unit=" patches",):
        # download stars and add to database
        download_stars.queue_to_write(download_degree2(t))

    download_stars.write_checkpoint()


@click.command()
@click.option('--ra', '-r', type=click.INT, default=0, help="The RA to start from.")
@click.option('--dec', '-d', type=click.INT, default=-90, help="The DEC to start from.")
@click.option(
    '--size',
    '-s',
    type=click.INT,
    default=360 * 180,
    help="The number of square degrees to download.",
)
def cli(ra, dec, size):
    main(start_ra=ra, start_dec=dec, size=size)


if __name__ == '__main__':
    cli()



#!.venv/bin/python3
'''
To be run after populating the database and before using it for queries.
'''

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

import config
from models import Source

engine = create_engine(
    f"sqlite:///{config.catalogue_db_path}", echo=True, pool_pre_ping=True
)

Session = sessionmaker(bind=engine)
session = Session()

# create indexes for Ra, Dec, Gmag
Source.create_indexes(engine)

# commit all wal changes
session.execute('PRAGMA wal_checkpoint;')
# exit wal journal mode
session.execute('PRAGMA journal_mode=delete;')



import operator
import re

import numpy as np

from collections import OrderedDict

from astropy import units as u
from astropy.table import Table

from sqlalchemy import create_engine, and_
from sqlalchemy.orm import sessionmaker

from gilmenel.exceptions import CatalogUnavailableError

from catalogue.models import Base, Source


catalog_url = 'sqlite://file::memory:'


def to_astrotable(data, names):
    '''Return an Astropy Table of data'''
    if len(data) == 0:
        return Table(names=names)  # empty Table

    # if data elements are of type Base used in the ORM model
    rows = [d._asdict() for d in data]
    table = Table(data=rows, names=names)

    return table


class Local:
    '''
    A mocked Vizier class to replace vquery.query_region

    Usage:
        query = Local(
            columns=["_r", 'Source', 'RA_ICRS', 'DE_ICRS', '+Gmag'],
            column_filters={
                "Gmag": (f"<=15"),
            },
            row_limit=max_stars)

        results = query.query_region(instr.target,
                                     radius=instr.instr_fov,
                                     catalog="I/345/gaia2")[0]
    '''

    def __init__(
        self, columns=None, column_filters={'Gmag': "<=10"}, row_limit=5, **kwargs,
    ):

        self.columns = columns  # not used
        self.column_filters = column_filters
        self.row_limit = row_limit

        engine = create_engine(catalog_url)

        Session = sessionmaker(bind=engine)  # noqa
        self.session = Session()

    def query_region(self, target, radius=1 * u.arcmin, catalog="I/345/gaia2"):

        if catalog != "I/345/gaia2":
            raise CatalogUnavailableError(
                f"Specified catalog '{catalog}' is not available"
            )

        comps = {
            "<": operator.lt,
            "<=": operator.le,
            "==": operator.eq,
            "!=": operator.ne,
            ">=": operator.ge,
            ">": operator.gt,
        }

        # transform to equatorial equivalent target
        equi_radius_ra = radius / np.cos(target.dec)

        ra_min = target.ra - equi_radius_ra
        ra_max = target.ra + equi_radius_ra

        dec_min = target.dec - radius
        dec_max = target.dec + radius

        # first do a box filter
        ra_filter = and_(
            Source.ra <= ra_max.to_value(u.deg), (Source.ra >= ra_min.to_value(u.deg))
        )
        dec_filter = and_(
            Source.dec <= dec_max.to_value(u.deg),
            (Source.dec >= dec_min.to_value(u.deg)),
        )

        # Gmag filter
        if 'Gmag' in self.column_filters:
            # remove comparators
            g_mag_number = re.sub(r'[!=<>]+', '', self.column_filters['Gmag'])
            # remove digits and points
            g_mag_comp = re.sub(r'[\d\.]+', '', self.column_filters['Gmag'])
            # limit queries by magnitude
            g_mag_filter = comps[g_mag_comp](Source.g_mag, g_mag_number)
        else:
            g_mag_filter = None

        query = self.session.query(Source)
        query = query.filter(ra_filter)
        query = query.filter(dec_filter)
        query = query.filter(g_mag_filter)
        query = query.limit(self.row_limit)

        results = query.all()

        # inject regular grid for testing
        # from tests.mocks import MockSource
        # results = []
        # for ra in np.linspace(ra_min, ra_max, num=10):
        #     ra = ra.to_value(u.deg)
        #     for dec in np.linspace(dec_min, dec_max, num=10):
        #         dec = dec.to_value(u.deg)
        #         source = MockSource(int(ra * 100 + dec), (
        #             ra,
        #             dec,
        #             None,
        #             float(g_mag_number),
        #         ))
        #         results.append(source)

        labels = OrderedDict(
            source_id='_id',
            source='DR2Name',
            ra='RA_ICRS',
            dec='DE_ICRS',
            g_mag='Gmag',
            rp_mag='RPmag',
        )

        # extract names of columns for correct ordering
        names = [k for k, _ in labels.items()]

        table = to_astrotable(results, names)

        # rename columns to match output from Vizier
        for old, new in labels.items():
            table.rename_column(old, new)

        # equator equivalent radius calculation
        target_ra = target.ra.to_value(u.deg)
        target_dec = target.dec.to_value(u.deg)
        candidates_ra = table['RA_ICRS']
        candidates_dec = table['DE_ICRS']

        ra_diff = target_ra - candidates_ra
        dec_scale = np.cos(np.radians(candidates_dec))

        # calculate radius in degrees
        table['_r'] = np.hypot(ra_diff * dec_scale, candidates_dec - target_dec)

        # degrees -> arcmin
        table['_r'] *= 60

        # then filter in place by radius
        mask = table['_r'] <= radius.to_value(u.arcmin)
        table = table[mask]

        # print(table)

        return [table]
